/******************************************************************************
*                                                                             *
*    calclib.c                                   Copyright(c) 2007 itow,y.    *
*                                                                             *
******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <float.h>
#include "calclib.h"


#define CL_PI		3.14159265358979323846
#define CL_RADIAN	(180.0/CL_PI)

#define DegToRad(x) ((x)*CL_PI/180.0)

#ifndef CL_USE_DOUBLE
#define MATH_FUNC(func) func##l
#else
#define MATH_FUNC(func) func
#endif

#ifndef UNICODE
#define CL_STR(str)	str
#define IsAlpha		isalpha
#define IsLower		islower
#define IsDigit		isdigit
#define IsXDigit	isxdigit
#define ToUpper		toupper
#define StrCmpI		strcmpi
#define SPrintF		sprintf
#else
#define CL_STR(str)	L##str
#define IsAlpha		iswalpha
#define IsLower		iswlower
#define IsDigit		iswdigit
#define IsXDigit	iswxdigit
#define ToUpper		towupper
#define StrCmpI		wcscmpi
#define SPrintF		swprintf
#endif

#define HexToInt(x) (IsDigit(x)?(x)-'0':ToUpper(x)-'A'+10)


static clError FunctionError;


static clError CalcAddSub(const clChar **ppExp,clValue *pAnswer);
static clError CalcMulDiv(const clChar **ppExp,clValue *pAnswer);
static clError CalcPower(const clChar **ppExp,clValue *pAnswer);
static clError CalcOther(const clChar **ppExp,clValue *pAnswer);
static clError CalcFuncVar(const clChar **ppExp,clValue *pAnswer);




/*
	w֐̗Onh
*/
#ifndef CL_USE_DOUBLE
int _matherrl(struct _exceptionl *x)
#else
int _matherr(struct _exception *x)
#endif
{
	switch (x->type) {
	case DOMAIN:
	case SING:
		FunctionError=CL_ERR_ARGUMENT;
		break;
	case OVERFLOW:
		FunctionError=CL_ERR_OVERFLOW;
		break;
	case UNDERFLOW:
		FunctionError=CL_ERR_UNDERFLOW;
		break;
	case TLOSS:
	case PLOSS:
		FunctionError=CL_ERR_LOSS;
		break;
	default:
		FunctionError=CL_ERR_UNKNOWN;
	}
	return 1;
}


/*
	Z/Z̏
*/
static clError CalcAddSub(const clChar **ppExp,clValue *pAnswer)
{
	const clChar *p;
	clValue n,n2;
	clChar c;
	clError Err;

	p=*ppExp;
	Err=CalcMulDiv(&p,&n);
	if (Err!=CL_ERR_OK)
		return Err;
	while (*p=='+' || *p=='-') {
		c=*p++;
		Err=CalcMulDiv(&p,&n2);
		if (Err!=CL_ERR_OK)
			return Err;
		if (c=='+') {
			/* Z */
			n+=n2;
		} else {
			/* Z */
			n-=n2;
		}
	}
	*ppExp=p;
	*pAnswer=n;
	return CL_ERR_OK;
}


/*
	Z/Z̏
*/
static clError CalcMulDiv(const clChar **ppExp,clValue *pAnswer)
{
	const clChar *p;
	clValue n,n2;
	clChar c;
	clError Err;

	p=*ppExp;
	Err=CalcPower(&p,&n);
	if (Err!=CL_ERR_OK)
		return Err;
	while (*p=='*' || *p=='/' || *p=='\\' || *p=='%') {
		c=*p++;
		Err=CalcPower(&p,&n2);
		if (Err!=CL_ERR_OK)
			return Err;
		switch (c) {
		case '*':
			/* Z */
			n*=n2;
			break;
		case '/':
			/* Z */
			if (n2==0.0)
				return CL_ERR_DIVIDEZERO;
			n/=n2;
			break;
		case '\\':
			/* Z(؂̂) */
			if (n2==0.0)
				return CL_ERR_DIVIDEZERO;
			n=MATH_FUNC(floor)(n/n2);
			break;
		case '%':
			/* ] */
			if (n2==0.0)
				return CL_ERR_DIVIDEZERO;
			n=MATH_FUNC(fmod)(n,n2);
			break;
		}
	}
	*ppExp=p;
	*pAnswer=n;
	return CL_ERR_OK;
}


/*
	ׂ̏
*/
static clError CalcPower(const clChar **ppExp,clValue *pAnswer)
{
	const clChar *p;
	clValue n,n2;
	clError Err;

	p=*ppExp;
	Err=CalcOther(&p,&n);
	if (Err!=CL_ERR_OK)
		return Err;
	while (*p=='^') {
		p++;
		Err=CalcOther(&p,&n2);
		if (Err!=CL_ERR_OK)
			return Err;
		n=MATH_FUNC(pow)(n,n2);
		if (FunctionError!=CL_ERR_OK)
			return FunctionError;
	}
	*ppExp=p;
	*pAnswer=n;
	return CL_ERR_OK;
}


/*
	16i̕𐔒lɕϊ
*/
static clError HexToValue(const clChar **ppHex,clValue *pValue)
{
	const clChar *p;
	clValue n;

	p=*ppHex;
	if (!IsXDigit(*p))
		return CL_ERR_SYNTAX;
	n=0.0;
	do {
		n*=16.0;
		n+=HexToInt(*p);
		p++;
	} while (IsXDigit(*p));
	*ppHex=p;
	*pValue=n;
	return CL_ERR_OK;
}


/*
	l//֐/萔̏
*/
static clError CalcOther(const clChar **ppExp,clValue *pAnswer)
{
	const clChar *p;
	clValue n;
	clError Err;

	p=*ppExp;
	while (*p==' ')
		p++;
	if (IsDigit(*p) || *p=='+' || *p=='-' || *p=='.') {
		/* l */
		if (*p=='0' && (*(p+1)=='x' || *(p+1)=='X')) {
			/* 16il */
			p+=2;
			Err=HexToValue(&p,&n);
			if (Err!=CL_ERR_OK)
				return Err;
		} else if (*p=='0' && (*(p+1)=='b' || *(p+1)=='B')) {
			/* 2il */
			p+=2;
			if (*p!='0' && *p!='1')
				return CL_ERR_SYNTAX;
			n=0.0;
			do {
				n*=2.0;
				if (*p=='1')
					n++;
				p++;
			} while (*p=='0' || *p=='1');
		} else {
			/* 10il */
#ifndef UNICODE
#ifndef CL_USE_DOUBLE
			n=_strtold(p,(char**)&p);
#else
			n=strtod(p,(char**)&p);
#endif
#else
#ifndef CL_USE_DOUBLE
			n=_wcstold(p,(wchar_t**)&p);
#else
			n=wcstod(p,(wchar_t**)&p);
#endif
#endif
			if (errno==ERANGE)
				return CL_ERR_OVERFLOW;
		}
	} else if (*p=='$') {
		/* 16il */
		p++;
		Err=HexToValue(&p,&n);
		if (Err!=CL_ERR_OK)
			return Err;
	} else if (*p=='(') {
		/*  */
		p++;
		Err=CalcAddSub(&p,&n);
		if (Err!=CL_ERR_OK)
			return Err;
		if (*p!=')')
			return CL_ERR_BRACKET;
		p++;
	} else if (IsAlpha(*p)) {
		/* ֐/萔 */
		Err=CalcFuncVar(&p,&n);
		if (Err!=CL_ERR_OK)
			return Err;
	} else
		return CL_ERR_SYNTAX;
	while (*p==' ')
		p++;
	*ppExp=p;
	*pAnswer=n;
	return CL_ERR_OK;
}


/*
	֐/萔̏
*/
static clError CalcFuncVar(const clChar **ppExp,clValue *pAnswer)
{
	const clChar *p;
	clChar Symbol[CL_MAX_SYMBOL];
	int i;
	clValue n;
	clError Err;

	p=*ppExp;
	i=0;
	while (i<CL_MAX_SYMBOL && (IsAlpha(*p) || *p=='_'))
		Symbol[i++]=*p++;
	if (i==CL_MAX_SYMBOL)
		return CL_ERR_TOOLARGESYMBOL;
	Symbol[i]='\0';
	if (*p=='(') {
		/* ֐ */
		p++;
		Err=CalcAddSub(&p,&n);
		if (Err!=CL_ERR_OK)
			return Err;
		if (*p!=')')
			return CL_ERR_SYNTAX;
		p++;
		if (StrCmpI(Symbol,CL_STR("abs"))==0) {
			n=MATH_FUNC(fabs)(n);
		} else if (StrCmpI(Symbol,CL_STR("acos"))==0) {
			if (n<-1.0 || n>1.0)
				return CL_ERR_ARGUMENT;
			n=MATH_FUNC(acos)(n)*CL_RADIAN;
		} else if (StrCmpI(Symbol,CL_STR("asin"))==0) {
			if (n<-1.0 || n>1.0)
				return CL_ERR_ARGUMENT;
			n=MATH_FUNC(asin)(n)*CL_RADIAN;
		} else if (StrCmpI(Symbol,CL_STR("atan"))==0) {
			n=MATH_FUNC(atan)(n)*CL_RADIAN;
		} else if (StrCmpI(Symbol,CL_STR("cos"))==0) {
			n=MATH_FUNC(cos)(DegToRad(n));
		} else if (StrCmpI(Symbol,CL_STR("cur"))==0) {
			clValue n2;

			n2=MATH_FUNC(fabs)(n);
			n=MATH_FUNC(pow)(n2,0.33333333333333333);
			if (n2<0.0)
				n=-n;
		} else if (StrCmpI(Symbol,CL_STR("exp"))==0) {
			n=MATH_FUNC(exp)(n);
		} else if (StrCmpI(Symbol,CL_STR("log"))==0) {
			if (n<=0.0)
				return CL_ERR_ARGUMENT;
			n=MATH_FUNC(log)(n);
		} else if (StrCmpI(Symbol,CL_STR("log10"))==0) {
			n=MATH_FUNC(log10)(n);
		} else if (StrCmpI(Symbol,CL_STR("sin"))==0) {
			n=MATH_FUNC(sin)(DegToRad(n));
		} else if (StrCmpI(Symbol,CL_STR("sqr"))==0
										|| StrCmpI(Symbol,CL_STR("sqrt"))==0) {
			if (n<0.0)
				return CL_ERR_ARGUMENT;
			n=MATH_FUNC(sqrt)(n);
		} else if (StrCmpI(Symbol,CL_STR("tan"))==0) {
			n=MATH_FUNC(tan)(DegToRad(n));
		} else
			return CL_ERR_UNKNOWNSYMBOL;
		if (FunctionError!=CL_ERR_OK)
			return FunctionError;
	} else {
		/* 萔 */
		if (StrCmpI(Symbol,CL_STR("pi"))==0
										|| StrCmpI(Symbol,CL_STR("pai"))==0)
			n=CL_PI;
		else
			return CL_ERR_UNKNOWNSYMBOL;
	}
	*ppExp=p;
	*pAnswer=n;
	return CL_ERR_OK;
}


/*
	vZs
*/
clError clCalc(const clChar *pExpression,clValue *pAnswer)
{
	const clChar *p;

	p=pExpression;
	FunctionError=CL_ERR_OK;
	return CalcAddSub(&p,pAnswer);
}


/*
	l10iɕϊ
*/
clError clFloatToDec(clValue Value,clChar *pString)
{
#if 0
	SPrintF(pString,
#ifndef CL_USE_DOUBLE
		CL_STR("%.*Lg"),
#else
		CL_STR("%.*g"),
#endif
		CL_MAX_DIGITS,Value);
#else
	clChar *pFirst,*pPoint,*pLast;
	int nDigits;

	pLast=pString+SPrintF(pString,
#ifndef CL_USE_DOUBLE
		CL_STR("%0.*Lf"),
#else
		CL_STR("%0.*f"),
#endif
		CL_MAX_DIGITS-1,Value);
	pFirst=pString;
	if (*pFirst=='-')
		pFirst++;
	for (pPoint=pFirst+1;*pPoint!='.';pPoint++);
	nDigits=pPoint-pFirst;
	if (nDigits>CL_MAX_DIGITS)
		return CL_ERR_OVERFLOW;
	if (nDigits<CL_MAX_DIGITS) {
		pLast=pString+SPrintF(pString,
#ifndef CL_USE_DOUBLE
			CL_STR("%0.*Lf"),
#else
			CL_STR("%0.*f"),
#endif
			CL_MAX_DIGITS-nDigits,Value);
		/* 0菜({͂ꂶ_ǁc) */
		while (*(pLast-1)=='0')
			pLast--;
		if (*(pLast-1)=='.')
			pLast--;
	} else {
		while (*(--pLast)=='0');
		if (*pLast!='.')
			pLast=pPoint+1;
	}
	*pLast='\0';
#endif
	return CL_ERR_OK;
}


/*
	l16iɕϊ
*/
clError clFloatToHex(clValue Value,clChar *pString,int nBits)
{
	clUInt64 n;

	n=(clUInt64)MATH_FUNC(floor)(Value);
	if (nBits==32 || (unsigned long)(n>>32)==0)
		SPrintF(pString,CL_STR("%lX"),(unsigned long)(n&0xFFFFFFFF));
	else
		SPrintF(pString,CL_STR("%lX%08lX"),
						(unsigned long)(n>>32),(unsigned long)(n&0xFFFFFFFF));
	return CL_ERR_OK;
}


/*
	l2iɕϊ
*/
clError clFloatToBin(clValue Value,clChar *pString,int nBits)
{
	clUInt64 n;
	unsigned long High,Low;
	clChar *p;
	int i;

	n=(clUInt64)MATH_FUNC(floor)(Value);
	High=(unsigned long)(n>>32);
	Low=(unsigned long)(n&0xFFFFFFFF);
	p=pString;
	if (nBits==64)
		for (i=0;i<32 && (High&(0x80000000>>i))==0;i++);
	if (nBits==32 || i==32) {
		for (i=0;i<32 && (Low&(0x80000000>>i))==0;i++);
		if (i==32) {
			*p++='0';
		} else {
			for (;i<32;i++)
				*p++=((Low>>(31-i))&1)+'0';
		}
	} else {
		for (;i<32;i++)
			*p++=((High>>(31-i))&1)+'0';
		for (i=0;i<32;i++)
			*p++=((Low>>(31-i))&1)+'0';
	}
	*p='\0';
	return CL_ERR_OK;
}
